import { ResolverOptions, DIDResolutionResult, ResolverRegistry, Resolver } from 'did-resolver';
import type { DagJWS, DIDProvider } from './types';
import { JWE } from './JWE';
export type { DIDResolutionResult } from 'did-resolver';
export * from './types';
export interface AuthenticateOptions {
    provider?: DIDProvider;
    aud?: string;
    paths?: Array<string>;
}
export interface AuthenticateParams {
    nonce: string;
    aud?: string;
    paths?: Array<string>;
}
export interface AuthenticateResponse extends AuthenticateParams {
    did: string;
    exp: number;
}
export interface CreateJWSOptions {
    did?: string;
    protected?: Record<string, any>;
    linkedBlock?: string;
}
export interface CreateJWSParams extends CreateJWSOptions {
    payload: any;
}
export interface CreateJWSResult {
    jws: DagJWS;
}
export interface VerifyJWSResult {
    kid: string;
    payload?: Record<string, any>;
    didResolutionResult: DIDResolutionResult;
}
export interface CreateJWEOptions {
    protectedHeader?: Record<string, any>;
    aad?: Uint8Array;
}
export interface DecryptJWEOptions {
    did?: string;
}
export interface DecryptJWEParams extends DecryptJWEOptions {
    jwe: JWE;
}
export interface DecryptJWEResult {
    cleartext: string;
}
export interface DagJWSResult {
    jws: DagJWS;
    linkedBlock: Uint8Array;
}
export interface DIDOptions {
    provider?: DIDProvider;
    resolver?: Resolver | ResolverRegistry;
    resolverOptions?: ResolverOptions;
}
/**
 * Interact with DIDs.
 */
export declare class DID {
    private _client?;
    private _id?;
    private _resolver;
    constructor({ provider, resolver, resolverOptions }?: DIDOptions);
    /**
     * Check if user is authenticated.
     */
    get authenticated(): boolean;
    /**
     * Get the DID identifier of the user.
     */
    get id(): string;
    /**
     * Set the DID provider of this instance.
     * Only callable if provider not already set.
     *
     * @param provider    The DIDProvider to use
     */
    setProvider(provider: DIDProvider): void;
    /**
     * Set the DID-resolver user by this instance
     *
     * @param resolver            Either a Resolver instance or an object with specific resolvers
     * @param resolverOptions     Options to use for the created resolver. Will be ignored if a Resolver instance is passed
     */
    setResolver(resolver: Resolver | ResolverRegistry, resolverOptions?: ResolverOptions): void;
    /**
     * Authenticate the user.
     */
    authenticate({ provider, paths, aud }?: AuthenticateOptions): Promise<string>;
    /**
     * Create a JWS encoded signature over the given payload.
     * Will be signed by the currently authenticated DID.
     *
     * @param payload             The payload to sign
     * @param options             Optional parameters
     */
    createJWS<T = any>(payload: T, options?: CreateJWSOptions): Promise<DagJWS>;
    /**
     * Create an IPFS compatibe DagJWS encoded signature over the given payload.
     * Will be signed by the currently authenticated DID.
     *
     * @param payload             The payload to sign, may include ipld links
     * @param options             Optional parameters
     */
    createDagJWS(payload: Record<string, any>, options?: CreateJWSOptions): Promise<DagJWSResult>;
    /**
     * Verify a JWS. Uses the 'kid' in the header as the way to resolve
     * the author public key.
     *
     * @param jws                 The JWS to verify
     * @returns                   Information about the signed JWS
     */
    verifyJWS(jws: string | DagJWS): Promise<VerifyJWSResult>;
    /**
     * Create a JWE encrypted to the given recipients.
     *
     * @param cleartext           The cleartext to be encrypted
     * @param recipients          An array of DIDs
     * @param options             Optional parameters
     */
    createJWE(cleartext: Uint8Array, recipients: Array<string>, options?: CreateJWEOptions): Promise<JWE>;
    /**
     * Create an IPFS compatibe DagJWE encrypted to the given recipients.
     *
     * @param cleartext           The cleartext to be encrypted, may include ipld links
     * @param recipients          An array of DIDs
     * @param options             Optional parameters
     */
    createDagJWE(cleartext: Record<string, any>, recipients: Array<string>, options?: CreateJWEOptions): Promise<JWE>;
    /**
     * Try to decrypt the given JWE with the currently authenticated user.
     *
     * @param jwe                 The JWE to decrypt
     * @param options             Optional parameters
     */
    decryptJWE(jwe: JWE, options?: DecryptJWEOptions): Promise<Uint8Array>;
    /**
     * Try to decrypt the given DagJWE with the currently authenticated user.
     *
     * @param jwe                 The JWE to decrypt
     * @param options             Optional parameters
     * @returns                   An ipld object
     */
    decryptDagJWE(jwe: JWE): Promise<Record<string, any>>;
    /**
     * Resolve the DID Document of the given DID.
     *
     * @param didUrl              The DID to resolve
     */
    resolve(didUrl: string): Promise<DIDResolutionResult>;
}
